import parse
from behave import register_type

TYPE_TO_PARSE_TYPE_MAP = {
    "int": "d",
    "float": "f",
}


class Word(str):
    pass


@parse.with_pattern(r"(.*)")
def parse_str(value: str) -> str:
    return value


@parse.with_pattern(r"\w+")
def parse_word(value: str) -> Word:
    return Word(value)


register_type(str=parse_str, Word=parse_word)


import inspect
from goat import string
from collections import OrderedDict

from behave import matchers, model as behave_model
from parse_type import cfparse

from goat import model
from goat.types import TYPE_TO_PARSE_TYPE_MAP


class GoatFormatter(string.Formatter):
    def __init__(self):
        self.unused_args = []

    def check_unused_args(self, used_args, args: tuple, kwargs: OrderedDict):
        kwarg_key_list = list(kwargs.keys())

        for used_arg in used_args:
            if isinstance(used_arg, int):
                key = kwarg_key_list[used_arg]
            else:
                key = used_arg
            kwargs.pop(key)

        self.unused_args = list(kwargs.keys())


class GoatMatcher(matchers.CFParseMatcher):
    def __init__(self, func, pattern, step_type=None):
        matchers.Matcher.__init__(self, func, pattern, step_type)
        self.context_params = []
        self.signature = inspect.signature(func)

        pattern = self.convert(pattern)
        self.parser = cfparse.Parser(pattern, self.custom_types)

    def match(self, step) -> behave_model.Match:
        result = self.check_match(step)
        if result is None:
            return None
        return model.Match(self.func, self.signature, result)

    def check_match(self, step) -> list:
        """Like matchers.CFParseMatcher.check_match but
        also add the implicit parameters from the context
        """
        args = []
        match = super().check_match(step)
        if match is None:
            return None

        for arg in match:
            args.append(model.Argument.from_argument(arg))

        for arg in self.context_params:
            args.append(model.Argument(0, 0, "", None, name=arg, implicit=True))

        return args

    def convert_type_to_parse_type(self, parameter):
        annotation = parameter.annotation
        if not isinstance(annotation, str):
            annotation = annotation.__name__
        annotation = TYPE_TO_PARSE_TYPE_MAP.get(annotation, annotation)
        return annotation

    def convert(self, pattern: str) -> str:
        """Convert the goat step string to CFParse String"""
        parameters = OrderedDict()
        for parameter in self.signature.parameters.values():
            annotation = self.convert_type_to_parse_type(parameter)
            parameters[parameter.name] = "{%s:%s}" % (parameter.name, annotation)

        formatter = GoatFormatter()

        # We have to use vformat here to ensure that kwargs will be OrderedDict
        values = parameters.values()
        parameter_list = list(values)
        converted_pattern = formatter.vformat(pattern, parameter_list, parameters)

        self.context_params = formatter.unused_args
        return converted_pattern


#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup



requirements = [
    "behave",
]

test_requirements = [
    "behave",
]

setup(
    name='goat',
    version='0.3',
    description='Goat implements a behave matcher which uses python3 function annotations for step definitiions',
    author='Ilja Bauer',
    author_email='i.bauer@cuescience.de',
    url='https://github.com/cuescience/goat',
    packages=[
        'goat',
    ],
    package_dir={'goat':
                 'goat'},
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='goat',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='features',
    tests_require=test_requirements
)



import inspect
from behave.model import Table, Text
from behave.runner import Context
from behave import model

CONTEXT_NAMESPACE = "_goat_{}"


class Argument(model.Argument):
    def __init__(self, *args, **kwargs):
        self.implicit = kwargs.pop("implicit", False)
        super().__init__(*args, **kwargs)

    @staticmethod
    def from_argument(argument: model.Argument, implicit=False):
        return Argument(
            argument.start,
            argument.end,
            argument.original,
            argument.value,
            name=argument.name,
            implicit=implicit
        )


class Match(model.Match):
    def __init__(self, func, signature, arguments: list=None):
        self.signature = signature
        super().__init__(func, arguments)

    @property
    def explicit_arguments(self):
        return filter(lambda x: not x.implicit, self.arguments)

    @property
    def implicit_arguments(self):
        return filter(lambda x: x.implicit, self.arguments)

    def run(self, context):
        """We have to overwrite this method because we don't want an implicit context
        """

        args = []
        kwargs = {}
        for arg in self.explicit_arguments:
            if arg.name is not None:
                kwargs[arg.name] = arg.value
            else:
                args.append(arg.value)

        for arg in self.implicit_arguments:
            if arg.name is not None:
                annotation = self.signature.parameters[arg.name].annotation
                annotation_name = annotation
                if not isinstance(annotation, str):
                    annotation_name = annotation.__name__

                if annotation is Table:
                    value = context.table
                elif annotation is Context:
                    value = context
                elif annotation is Text:
                    value = context.text
                else:
                    value = context.__getattr__(CONTEXT_NAMESPACE.format(annotation_name))

                kwargs[arg.name] = value
            else:
                raise RuntimeError("Argument name shouldn't be None")

        with context.user_mode():
            return_value = self.func(*args, **kwargs)
            return_annotation = self.signature.return_annotation
            if return_annotation == inspect.Signature.empty:
                return

            if not isinstance(return_annotation, str):
                return_annotation = return_annotation.__name__

            context.__setattr__(CONTEXT_NAMESPACE.format(return_annotation), return_value)